@@ -202,6 +202,11 @@ extern "C" {
     /// know for certain.
     pub fn mi_zalloc_small(size: usize) -> *mut c_void;
 
+    /// Return the available bytes in a memory block.
+    ///
+    /// The returned size can be used to call `mi_expand` successfully.
+    pub fn mi_usable_size(p: *const c_void) -> usize;
+
     /// Return the used allocation size.
     ///
     /// Returns the size `n` that will be allocated, where `n >= size`.
@@ -465,29 +470,26 @@ pub type mi_error_fun = Option<unsafe extern "C" fn(code: c_int, arg: *mut c_voi
 /// Runtime options. All options are false by default.
 pub type mi_option_t = c_int;
 
+#[cfg(feature = "arena")]
+/// Arena Id
+pub type mi_arena_id_t = c_int;
+
 // Note: mimalloc doc website seems to have the order of show_stats and
 // show_errors reversed as of 1.6.3, however what I have here is correct:
 // https://github.com/microsoft/mimalloc/issues/266#issuecomment-653822341
 
-/// Option allowing printing error messages to stderr.
+/// Print error messages to `stderr`.
 pub const mi_option_show_errors: mi_option_t = 0;
 
-/// Option allowing printing statistics to stderr when the program is done.
+/// Print statistics to `stderr` when the program is done.
 pub const mi_option_show_stats: mi_option_t = 1;
 
-/// Option allowing printing verbose messages to stderr.
+/// Print verbose messages to `stderr`.
 pub const mi_option_verbose: mi_option_t = 2;
 
-/// Option (experimental) specifying eagerly commit segments (4MiB) (enabled by default).
-pub const mi_option_eager_commit: mi_option_t = 3;
-
-/// Option (experimental) specifying eagerly commit large (256MiB) memory regions (enabled by default, except on Windows).
-pub const mi_option_eager_region_commit: mi_option_t = 4;
-
-/// Experimental
-pub const mi_option_reset_decommits: mi_option_t = 5;
-
-/// Option (experimental) to use large OS pages (2MiB in size) if possible.
+/// ### The following options are experimental
+///
+/// Option (experimental) Use large OS pages (2MiB in size) if possible.
 ///
 /// Use large OS pages (2MiB) when available; for some workloads this can
 /// significantly improve performance. Use mi_option_verbose to check if
@@ -499,7 +501,7 @@ pub const mi_option_reset_decommits: mi_option_t = 5;
 /// instead whenever possible).
 pub const mi_option_large_os_pages: mi_option_t = 6;
 
-/// Option (experimental) specifying number of huge OS pages (1GiB in size) to reserve at the start of the program.
+/// Option (experimental) The number of huge OS pages (1GiB in size) to reserve at the start of the program.
 ///
 /// This reserves the huge pages at startup and sometimes this can give a large (latency) performance
 /// improvement on big workloads. Usually it is better to not use MIMALLOC_LARGE_OS_PAGES in
@@ -512,56 +514,39 @@ pub const mi_option_large_os_pages: mi_option_t = 6;
 /// allocate just a little to take up space in the huge OS page area (which cannot be reset).
 pub const mi_option_reserve_huge_os_pages: mi_option_t = 7;
 
-/// TODO: update later
-pub const mi_option_reserve_os_memory: mi_option_t = 8;
-
-/// Option (experimental) specifying number of segments per thread to keep cached.
-pub const mi_option_segment_cache: mi_option_t = 9;
-
-/// Option (experimental) to reset page memory after mi_option_reset_delay milliseconds when it becomes free.
+/// Option (experimental) Reserve huge OS pages at node N.
 ///
-/// By default, mimalloc will reset (or purge) OS pages that are not in use, to signal to the OS
-/// that the underlying physical memory can be reused. This can reduce memory fragmentation in
-/// long running (server) programs. By setting it to 0 this will no longer be done which can improve
-/// performance for batch-like programs. As an alternative, the mi_option_reset_delay= can be set
-/// higher (100ms by default) to make the page reset occur less frequently instead of turning it
-/// off completely.
-///
-/// Default: 1 (true)
-pub const mi_option_page_reset: mi_option_t = 10;
+/// The huge pages are usually allocated evenly among NUMA nodes.
+/// You can use mi_option_reserve_huge_os_pages_at=N where `N` is the numa node (starting at 0) to allocate all
+/// the huge pages at a specific numa node instead.
+pub const mi_option_reserve_huge_os_pages_at: mi_option_t = 8;
 
-/// Experimental
-pub const mi_option_abandoned_page_reset: mi_option_t = 11;
+/// Option (experimental) Reserve specified amount of OS memory at startup, e.g. "1g" or "512m".
+pub const mi_option_reserve_os_memory: mi_option_t = 9;
 
-/// Experimental
-pub const mi_option_segment_reset: mi_option_t = 12;
+/// Option (experimental) the first N segments per thread are not eagerly committed (=1).
+pub const mi_option_eager_commit_delay: mi_option_t = 14;
 
-/// Experimental
-pub const mi_option_eager_commit_delay: mi_option_t = 13;
+/// Option (experimental) Pretend there are at most N NUMA nodes; Use 0 to use the actual detected NUMA nodes at runtime.
+pub const mi_option_use_numa_nodes: mi_option_t = 16;
 
-/// Option (experimental) specifying delay in milli-seconds before resetting a page (100ms by default).
-pub const mi_option_reset_delay: mi_option_t = 14;
+/// Option (experimental) If set to 1, do not use OS memory for allocation (but only pre-reserved arenas)
+pub const mi_option_limit_os_alloc: mi_option_t = 17;
 
-/// Option (experimental) to pretend there are at most N NUMA nodes.
-///
-/// If not set, the actual NUMA nodes are detected at runtime. Setting N to 1 may avoid
-/// problems in some virtual environments. Also, setting it to a lower number than the
-/// actual NUMA nodes is fine and will only cause threads to potentially allocate more
-/// memory across actual NUMA nodes (but this can happen in any case as NUMA local
-/// allocation is always a best effort but not guaranteed).
-pub const mi_option_use_numa_nodes: mi_option_t = 15;
+/// Option (experimental) OS tag to assign to mimalloc'd memory
+pub const mi_option_os_tag: mi_option_t = 18;
 
-/// TODO: update later
-pub const mi_option_limit_os_alloc: mi_option_t = 16;
+/// Option (experimental)
+pub const mi_option_max_errors: mi_option_t = 19;
 
-/// Option (experimental) specifying OS tag to assign to mimalloc'd memory.
-pub const mi_option_os_tag: mi_option_t = 17;
+/// Option (experimental)
+pub const mi_option_max_warnings: mi_option_t = 20;
 
-/// Experimental
-pub const mi_option_max_errors: mi_option_t = 18;
+/// Option (experimental)
+pub const mi_option_max_segment_reclaim: mi_option_t = 21;
 
-/// Experimental
-pub const mi_option_max_warnings: mi_option_t = 19;
+/// Last option.
+pub const _mi_option_last: mi_option_t = 37;
 
 extern "C" {
     // Note: mi_option_{enable,disable} aren't exposed because they're redundant
@@ -670,6 +655,10 @@ pub struct mi_heap_area_t {
     pub used: usize,
     /// Size in bytes of one block.
     pub block_size: usize,
+    /// Size in bytes of a full block including padding and metadata.
+    pub full_block_size: usize,
+    /// Heap tag associated with this area
+    pub heap_tag: i32,
 }
 
 /// Visitor function passed to [`mi_heap_visit_blocks`]
@@ -963,6 +952,63 @@ extern "C" {
         visitor: mi_block_visit_fun,
         arg: *mut c_void,
     ) -> bool;
+
+    #[cfg(feature = "arena")]
+    /// Create a heap that only allocates in the specified arena
+    pub fn mi_heap_new_in_arena(arena_id: mi_arena_id_t) -> *mut mi_heap_t;
+
+    #[cfg(feature = "arena")]
+    /// Reserve OS memory for use by mimalloc. Reserved areas are used
+    /// before allocating from the OS again. By reserving a large area upfront,
+    /// allocation can be more efficient, and can be better managed on systems
+    /// without `mmap`/`VirtualAlloc` (like WASM for example).
+    ///
+    /// - `size` The size to reserve.
+    /// - `commit` Commit the memory upfront.
+    /// - `allow_large` Allow large OS pages (2MiB) to be used?
+    /// - `exclusive` Only allow allocations if specifically for this arena.
+    /// - `arena_id` Pointer who's value will be set to the new arena_id if successful.
+    ///
+    /// Returns 0 if successful, and an error code otherwise (e.g. `ENOMEM`)
+    pub fn mi_reserve_os_memory_ex(
+        size: usize,
+        commit: bool,
+        allow_large: bool,
+        exclusive: bool,
+        arena_id: *mut mi_arena_id_t,
+    ) -> c_int;
+
+    #[cfg(feature = "arena")]
+    /// Manage a particular memory area for use by mimalloc.
+    /// This is just like `mi_reserve_os_memory_ex` except that the area should already be
+    /// allocated in some manner and available for use my mimalloc.
+    ///
+    /// # Safety
+    /// mimalloc will likely segfault when allocating from the arena if the arena `start` & `size`
+    /// aren't aligned with mimalloc's `MI_SEGMENT_ALIGN` (e.g. 32MB on x86_64 machines).
+    ///
+    /// - `start` Start of the memory area
+    /// - `size` The size of the memory area. Must be large than `MI_ARENA_BLOCK_SIZE` (e.g. 64MB
+    ///          on x86_64 machines).
+    /// - `commit` Set true if the memory range is already commited.
+    /// - `is_large` Set true if the memory range consists of large files, or if the memory should
+    ///              not be decommitted or protected (like rdma etc.).
+    /// - `is_zero` Set true if the memory range consists only of zeros.
+    /// - `numa_node` Possible associated numa node or `-1`.
+    /// - `exclusive` Only allow allocations if specifically for this arena.
+    /// - `arena_id` Pointer who's value will be set to the new arena_id if successful.
+    ///
+    /// Returns `true` if arena was successfully allocated
+    pub fn mi_manage_os_memory_ex(
+        start: *const c_void,
+        size: usize,
+        is_committed: bool,
+        is_large: bool,
+        is_zero: bool,
+        numa_node: c_int,
+        exclusive: bool,
+        arena_id: *mut mi_arena_id_t,
+    ) -> bool;
 }
 
 #[cfg(test)]
@@ -970,17 +1016,29 @@ mod tests {
     use super::*;
 
     #[test]
-    fn runtime_option_page_reset() {
-        unsafe {
-            // page reset
-            assert_eq!(mi_option_get(mi_option_page_reset), 1);
-            mi_option_set(mi_option_page_reset, 2);
-            assert_eq!(mi_option_get(mi_option_page_reset), 2);
+    fn it_calculates_usable_size() {
+        let ptr = unsafe { mi_malloc(32) } as *mut u8;
+        let usable_size = unsafe { mi_usable_size(ptr as *mut c_void) };
+        assert!(
+            usable_size >= 32,
+            "usable_size should at least equal to the allocated size"
+        );
+    }
 
-            // page reset delay
-            assert_eq!(mi_option_get(mi_option_reset_delay), 100);
-            mi_option_set(mi_option_reset_delay, 10_000);
-            assert_eq!(mi_option_get(mi_option_reset_delay), 10_000);
+    #[test]
+    fn runtime_stable_option() {
+        unsafe {
+            assert_eq!(mi_option_get(mi_option_show_errors), 0);
+            mi_option_set(mi_option_show_errors, 1);
+            assert_eq!(mi_option_get(mi_option_show_errors), 1);
+
+            assert_eq!(mi_option_get(mi_option_show_stats), 0);
+            mi_option_set(mi_option_show_stats, 1);
+            assert_eq!(mi_option_get(mi_option_show_stats), 1);
+
+            assert_eq!(mi_option_get(mi_option_verbose), 0);
+            mi_option_set(mi_option_verbose, 1);
+            assert_eq!(mi_option_get(mi_option_verbose), 1);
         }
     }
 }
